



pcreate --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pserve --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pshell --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
proutes --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pviews --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
ptweens --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
prequest --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pdistreport --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
